@@ -8,5 +8,7 @@
 %% API.
 
 start() ->
+	ok = application:start(crypto),
+	ok = application:start(ranch),
 	ok = application:start(cowboy),
 	ok = application:start(hello_world).
