@@ -5,14 +5,14 @@
 
 -export([init/3]).
 -export([handle/2]).
--export([terminate/2]).
+-export([terminate/3]).
 
 init(_Transport, Req, []) ->
 	{ok, Req, undefined}.
 
 handle(Req, State) ->
-	{ok, Req2} = cowboy_http_req:reply(200, [], <<"Hello world!">>, Req),
+	{ok, Req2} = cowboy_req:reply(200, [], <<"Hello world!">>, Req),
 	{ok, Req2, State}.
 
-terminate(_Req, _State) ->
+terminate(_Reason, _Req, _State) ->
 	ok.
